from __future__ import annotations
import SOMcreator
from SOMcreator import classes
from SOMcreator.filehandling import core
from SOMcreator.filehandling.constants import VALUE, VALUE_TYPE, DATA_TYPE, CHILD_INHERITS_VALUE, REVIT_MAPPING
from SOMcreator.constants.value_constants import OLD_DATATYPE_DICT
from SOMcreator import filehandling
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from SOMcreator.filehandling.typing import AttributeDict


def load(proj: SOMcreator.Project, attribute_dict: dict, identifier: str,
         property_set: classes.PropertySet, ) -> None:
    name, description, optional, parent, filter_matrix = core.get_basics(proj, attribute_dict)
    value = attribute_dict[VALUE]
    value_type = attribute_dict[VALUE_TYPE]
    data_type = attribute_dict[DATA_TYPE]

    # compatibility for Datatype import that uses XML-Datatypes such as xs:string
    if data_type in OLD_DATATYPE_DICT:
        data_type = OLD_DATATYPE_DICT[data_type]

    child_inherits_value = attribute_dict[CHILD_INHERITS_VALUE]
    revit_mapping = attribute_dict[REVIT_MAPPING]
    attribute = classes.Attribute(property_set=property_set, name=name, value=value, value_type=value_type,
                                  data_type=data_type,
                                  child_inherits_values=child_inherits_value, uuid=identifier,
                                  description=description, optional=optional, revit_mapping=revit_mapping,
                                  project=proj, filter_matrix=filter_matrix)
    filehandling.parent_dict[attribute] = parent


def write(attribute: classes.Attribute) -> AttributeDict:
    attribute_dict: AttributeDict = dict()
    core.write_basics(attribute_dict, attribute)
    attribute_dict[DATA_TYPE] = attribute.data_type
    attribute_dict[VALUE_TYPE] = attribute.value_type
    attribute_dict[CHILD_INHERITS_VALUE] = attribute.child_inherits_values
    attribute_dict[REVIT_MAPPING] = attribute.revit_name
    attribute_dict[VALUE] = attribute.value
    return attribute_dict



from ..filehandling.constants import *
NODE = "Node"
NODES = "Nodes"
IS_ROOT = "root"
AGGREGATES_TO = "aggregates_to"
SCRIPTS = "Scripts"
SCRIPT = "Script"
IS_CONCEPT = "is_concept"
IS_IDENTIFIER = "is_identifier"
NONE = "None"
IDENTIFIER = "identifer"
AGGREGATE = "Aggregate"
ATTRIBUTE = "Attribute"
PROPERTY_SET = "PropertySet"
PREDEFINED_PSET = "PredefinedPropertySet"
PREDEFINED_PROPERTY_WINDOW_NAME = "Predefined Properties"
INHERITED_TEXT = "Predefined Pset"



import SOMcreator
from SOMcreator.classes import Project, UseCase, Phase, Hirarchy


def _calculate_new_filter_matrix(filter_matrix, existing_project: Project, import_project: Project, item: Hirarchy,
                                 phase_mapping, use_case_mapping):
    for import_phase in import_project.get_project_phase_list():
        existing_phase = phase_mapping.get(import_phase)
        if existing_phase is None:
            continue

        for import_use_case in import_project.get_use_case_list():
            existing_use_case = use_case_mapping.get(import_use_case)
            if existing_use_case is None:
                continue
            phase_index = existing_project.get_phase_index(existing_phase)
            use_case_index = existing_project.get_use_case_index(existing_use_case)
            import_state = item.get_filter_state(import_phase, import_use_case)
            filter_matrix[phase_index][use_case_index] = import_state


def _merge_dicts(d1: dict, d2: dict):
    for key, value2 in d2.items():
        value1 = d1.get(key)
        if isinstance(value1, dict) and isinstance(value2, dict):
            _merge_dicts(value1, value2)
        elif value1 is None:
            d1[key] = value2
    return d1


def _add_item(existing_project: SOMcreator.Project, import_project: SOMcreator.Project, item, phase_mapping,
              use_case_mapping):
    new_filter_matrix = existing_project.create_filter_matrix(True)
    _calculate_new_filter_matrix(new_filter_matrix, existing_project, import_project, item, phase_mapping,
                                 use_case_mapping)
    existing_project.add_item(item)
    item._project = existing_project
    item._filter_matrix = new_filter_matrix


def _import_object(existing_project, import_project, obj, old_predefined_psets_mapping, phase_mapping,
                   use_case_mapping):
    _add_item(existing_project, import_project, obj, phase_mapping, use_case_mapping)
    for property_set in obj.get_all_property_sets():
        _import_pset(existing_project, import_project, property_set, old_predefined_psets_mapping, phase_mapping,
                     use_case_mapping)

    for aggregation in obj.aggregations:
        _add_item(existing_project, import_project, aggregation, phase_mapping, use_case_mapping)


def _import_pset(existing_project, import_project, property_set, old_predefined_psets_mapping, phase_mapping,
                 use_case_mapping):
    parent = old_predefined_psets_mapping.get(property_set.parent)
    _add_item(existing_project, import_project, property_set, phase_mapping, use_case_mapping)

    if parent is not None:
        property_set.parent = parent
        for attribute in property_set.get_all_attributes():
            _import_attribute(existing_project, import_project, attribute, phase_mapping, use_case_mapping, parent)
    else:
        for attribute in property_set.get_all_attributes():
            _import_attribute(existing_project, import_project, attribute, phase_mapping, use_case_mapping)


def _import_attribute(existing_project, import_project, attribute, phase_mapping, use_case_mapping,
                      parent_pset: SOMcreator.PropertySet = None):
    if parent_pset:
        parent_attribute = {a.name: a for a in parent_pset.attributes}.get(attribute.name)
        if parent_attribute:
            attribute.parent = parent_attribute
    _add_item(existing_project, import_project, attribute, phase_mapping, use_case_mapping)


def merge_projects(existing_project: Project, import_project: Project, phase_mapping: dict[Phase, Phase],
                   use_case_mapping: dict[UseCase, UseCase]):
    """

    :param existing_project: Existing Project
    :param import_project: Project that will be imported into existing Project
    :param phase_mapping: mapping Dict import Phase as Key existing Phase as Value
    :param use_case_mapping: mapping Dict import UseCase as Key existing UseCase as Value
    :return:
    """
    existing_identifiers = {o.ident_value for o in existing_project.get_all_objects()}
    import_predef_pset_dict = {p: p for p in import_project.get_predefined_psets()}
    existing_predefined_pset_name_dict = {p.name: p for p in existing_project.get_predefined_psets()}

    for import_predef_pset in import_predef_pset_dict.keys():
        new_pset = existing_predefined_pset_name_dict.get(import_predef_pset.name)
        if not new_pset:
            _add_item(existing_project, import_project, import_predef_pset, phase_mapping, use_case_mapping)
        else:
            import_predef_pset_dict[import_predef_pset] = new_pset

    for obj in import_project.get_all_objects():
        if obj.ident_value not in existing_identifiers:
            _import_object(existing_project, import_project, obj, import_predef_pset_dict, phase_mapping,
                           use_case_mapping)

    existing_project.plugin_dict = _merge_dicts(existing_project.plugin_dict, import_project.plugin_dict)
    existing_project.import_dict = _merge_dicts(existing_project.import_dict, import_project.import_dict)



from __future__ import annotations
from SOMcreator.filehandling import core
from SOMcreator.filehandling.constants import PREDEFINED_PSETS
from SOMcreator.filehandling import property_set
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from SOMcreator.filehandling.typing import MainDict
    from SOMcreator import Project


def load(project: Project, main_dict: MainDict):
    predef_pset_dict = main_dict.get(PREDEFINED_PSETS)
    core.remove_part_of_dict(PREDEFINED_PSETS)
    predef_pset_dict = dict() if core.check_dict(predef_pset_dict, PREDEFINED_PSETS) else predef_pset_dict

    for uuid_ident, entity_dict in predef_pset_dict.items():
        property_set.load(project, entity_dict, uuid_ident, None)


def write(proj: Project, main_dict: MainDict):
    main_dict[PREDEFINED_PSETS] = dict()
    for predefined_property_set in sorted(proj.get_predefined_psets(), key=lambda x: x.uuid):
        main_dict[PREDEFINED_PSETS][predefined_property_set.uuid] = property_set.write_entry(predefined_property_set)



from __future__ import annotations
from typing import TYPE_CHECKING, Type

from .constants import PROJECT, NAME, AUTHOR, VERSION, AGGREGATION_PSET, AGGREGATION_ATTRIBUTE, CURRENT_PR0JECT_PHASE, \
    CURRENT_USE_CASE, FILTER_MATRIX, PROJECT_PHASES, USE_CASES
from .typing import ProjectDict, FilterDict, MainDict
import SOMcreator
from SOMcreator import classes
from SOMcreator.filehandling import core

if TYPE_CHECKING:
    from SOMcreator import Project


def _load_filter_matrix(project_dict: ProjectDict, use_case_list: list[classes.UseCase],
                        phase_list: list[classes.Phase]):
    old_filter_matrix: list[list[bool]] = project_dict.get(FILTER_MATRIX)
    filter_matrix = list()
    for _ in phase_list:
        filter_matrix.append([True for __ in use_case_list])

    if old_filter_matrix is None:
        return filter_matrix

    for old_phase, new_phase in zip(old_filter_matrix, filter_matrix):
        for index, old_value in enumerate(old_phase):
            new_phase[index] = old_value

    return filter_matrix


def _load_filter(current_state: str | int | None,
                 value_list: list[classes.UseCase] | list[classes.Phase]):
    if isinstance(current_state, str):
        current_state = {value.name: value for value in value_list}.get(current_state)
    elif isinstance(current_state, int):
        current_state = value_list[current_state]
    if current_state is None:
        current_state = value_list[0]

    return current_state, value_list


def _load_usecases(project_dict: ProjectDict) -> tuple[classes.UseCase, list[classes.UseCase]]:
    current_use_case = project_dict.get(CURRENT_USE_CASE)
    use_case_list = SOMcreator.filehandling.use_case_list
    return _load_filter(current_use_case, use_case_list)


def _load_phases(project_dict: ProjectDict) -> tuple[classes.Phase, list[classes.Phase]]:
    current_project_phase = project_dict.get(CURRENT_PR0JECT_PHASE)
    phase_list = SOMcreator.filehandling.phase_list
    return _load_filter(current_project_phase, phase_list)


def load(cls: Type[Project], main_dict: MainDict) -> tuple[Project, dict]:
    project_dict: ProjectDict = main_dict.get(PROJECT)
    core.remove_part_of_dict(PROJECT)

    name = project_dict.get(NAME)
    author = project_dict.get(AUTHOR)
    version = project_dict.get(VERSION)

    aggregation_pset_name = project_dict.get(AGGREGATION_PSET)
    aggregation_attribute = project_dict.get(AGGREGATION_ATTRIBUTE)

    current_use_case, use_case_list = _load_usecases(project_dict)
    current_phase, phase_list = _load_phases(project_dict)
    filter_matrix = _load_filter_matrix(project_dict, use_case_list, phase_list)

    proj = cls(name, author, phase_list, use_case_list, filter_matrix)
    proj.version = version
    if aggregation_pset_name is not None:
        proj.aggregation_pset = aggregation_pset_name
    if aggregation_attribute is not None:
        proj.aggregation_attribute = aggregation_attribute

    return proj, project_dict


def write(project, main_dict: MainDict) -> None:
    main_dict[PROJECT] = dict()
    project_dict: ProjectDict = main_dict[PROJECT]
    project_dict[NAME] = project.name
    project_dict[AUTHOR] = project.author
    project_dict[VERSION] = project.version
    project_dict[AGGREGATION_ATTRIBUTE] = project.aggregation_attribute
    project_dict[AGGREGATION_PSET] = project.aggregation_pset
    project_dict[CURRENT_PR0JECT_PHASE] = project.get_project_phase_list().index(project.current_project_phase)
    project_dict[CURRENT_USE_CASE] = project.get_use_case_list().index(project.current_use_case)
    project_dict[PROJECT_PHASES] = _write_filter_dict(project.get_project_phase_list())
    project_dict[USE_CASES] = _write_filter_dict(project.get_use_case_list())
    project_dict[FILTER_MATRIX] = project.get_filter_matrix()


def _write_filter_dict(filter_list: list[classes.Phase] | list[classes.UseCase]) -> list[FilterDict]:
    fl = list()
    for fil in filter_list:
        fl.append({
            "name":        fil.name,
            "long_name":   fil.long_name,
            "description": fil.description
        })
    return fl



from __future__ import annotations
from typing import TYPE_CHECKING
from SOMcreator.filehandling.constants import IFC_MAPPINGS, ABBREVIATION, PROPERTY_SETS, IDENT_ATTRIBUTE, OBJECTS
from SOMcreator.filehandling import property_set
import SOMcreator
from SOMcreator import classes
from SOMcreator.filehandling import core

if TYPE_CHECKING:
    from SOMcreator import Project
    from SOMcreator.filehandling.typing import ObjectDict, MainDict


### Import ###

def _load_object(proj, object_dict: ObjectDict, identifier: str) -> None:
    name, description, optional, parent, filter_matrix = core.get_basics(proj, object_dict)
    ifc_mapping = object_dict[IFC_MAPPINGS]
    if isinstance(ifc_mapping, list):
        ifc_mapping = set(ifc_mapping)

    abbreviation = object_dict.get(ABBREVIATION)

    obj = classes.Object(name=name, ident_attrib=None, uuid=identifier, ifc_mapping=ifc_mapping,
                         description=description, optional=optional, abbreviation=abbreviation, project=proj,
                         filter_matrix=filter_matrix)
    property_sets_dict = object_dict[PROPERTY_SETS]
    for ident, pset_dict in property_sets_dict.items():
        property_set.load(proj, pset_dict, ident, obj)
    ident_attrib_id = object_dict[IDENT_ATTRIBUTE]
    ident_attrib = classes.get_element_by_uuid(ident_attrib_id)
    obj.ident_attrib = ident_attrib
    SOMcreator.filehandling.parent_dict[obj] = parent


def load(proj: Project, main_dict: dict):
    objects_dict: dict[str, ObjectDict] = main_dict.get(OBJECTS)
    core.remove_part_of_dict(OBJECTS)

    objects_dict = dict() if core.check_dict(objects_dict, OBJECTS) else objects_dict

    for uuid_ident, entity_dict in objects_dict.items():
        _load_object(proj, entity_dict, uuid_ident)


### Export ###
def _write_object(element: classes.Object) -> ObjectDict:
    object_dict: ObjectDict = dict()
    core.write_basics(object_dict, element)

    if isinstance(element.ifc_mapping, set):
        object_dict[IFC_MAPPINGS] = list(element.ifc_mapping)
    else:
        object_dict[IFC_MAPPINGS] = list(element.ifc_mapping)

    psets_dict = dict()
    for pset in element.get_all_property_sets():
        psets_dict[pset.uuid] = property_set.write_entry(pset)

    object_dict[PROPERTY_SETS] = psets_dict
    object_dict[ABBREVIATION] = element.abbreviation

    if isinstance(element.ident_attrib, classes.Attribute):
        object_dict[IDENT_ATTRIBUTE] = element.ident_attrib.uuid
    else:
        object_dict[IDENT_ATTRIBUTE] = element.ident_attrib

    return object_dict


def write(proj: Project, main_dict: MainDict):
    main_dict[OBJECTS] = dict()

    for obj in sorted(proj.get_all_objects(), key=lambda o: o.uuid):
        main_dict[OBJECTS][obj.uuid] = _write_object(obj)



import logging

from openpyxl import Workbook
from openpyxl.worksheet.worksheet import Worksheet

from .. import classes, constants
from ..constants import value_constants

TITLES = ["Definition", "Zuweisung", "Mapping"]
COLUMNS = ["AttributeName",
           "AttributeTyp",
           "AttributeValue",
           "AttributMin",
           "AttributMax",
           "AttrEinh",
           "AttrEingab",
           "AttVorgabe_I",
           "AttVorgabe_II",
           "AttVorgabe_III",
           "AttVorgabe_IV"]

INTERNAL_COLUMNS = ["Objekt", "AttributAllplan", "AttributIfc", "Pset", "Type"]


def create_mapping(project: classes.Project, path: str, allplan_mapping_name: str):
    def transform_datatype(data_type: str) -> str:
        if data_type == value_constants.INTEGER:
            return "Ganzzahl"
        if data_type == value_constants.REAL:
            return "Fließkommazahl"
        return "Text"

    def create_definition(worksheet: Worksheet) -> dict[str, str]:

        for x, text in enumerate(COLUMNS):
            worksheet.cell(row=1, column=x + 1, value=text)

        attribute_dict: dict[str, str] = dict()

        for attribute in classes.Attribute:
            data_type = attribute.data_type

            if attribute.name in attribute_dict:
                if data_type != attribute_dict[attribute.name]:
                    logging.warning(f"Achtung bei {attribute.property_set.object.name} -> {attribute.property_set.name}"
                                    f":{attribute.name} neuer Datentyp: {data_type} "
                                    f" alter Datentyp: {attribute_dict[attribute.name]}")
            else:
                attribute_dict[attribute.name] = data_type

        for row_index, [name, data_type] in enumerate(attribute_dict.items()):
            row = 2 + row_index
            worksheet.cell(row=row, column=1, value=name)
            worksheet.cell(row=row, column=2, value=transform_datatype(data_type))
            if data_type == value_constants.BOOLEAN:
                worksheet.cell(row=row, column=7, value="CheckBox")
        return attribute_dict

    def create_zuweisung(kenner: str, worksheet: Worksheet):

        def get_attrib_count(obj: classes.Object):
            return sum(len([attrib for attrib in pset.attributes]) for pset in obj.property_sets)

        max_attribs = max(
            get_attrib_count(obj) for obj in project.objects)
        header = ["Kenner"] + ["Wert", "Name"] * max_attribs
        [worksheet.cell(1, i + 1, text) for i, text in enumerate(header)]   #print Header
        worksheet.cell(2, 1, kenner)
        row_index = 2
        for obj in project.objects:
            worksheet.cell(row_index, 2, obj.ident_value)
            col_index = 3
            for propery_set in obj.property_sets:
                for attribute in propery_set.attributes:
                    if attribute.name != kenner:
                        worksheet.cell(row_index, col_index, attribute.name)
                        col_index += 2

            row_index += 1

    def create_internal_mapping(attribute_dict: dict[str, str], worksheet: Worksheet):
        def transform_type(t: str) -> str:
            if t == value_constants.INTEGER:
                return "IfcInteger"
            if t == value_constants.REAL:
                return "IfcReal"
            if t == value_constants.BOOLEAN:
                return "IfcBoolean"
            return "IfcLabel"

        for x, text in enumerate(INTERNAL_COLUMNS):
            worksheet.cell(row=1, column=x + 1, value=text)
        worksheet.cell(2, 1, "All")
        for row_index, (attribute_name, attribute_datatype) in enumerate(sorted(attribute_dict.items())):
            worksheet.cell(2 + row_index, 2, attribute_name)
            worksheet.cell(2 + row_index, 4, allplan_mapping_name)
            worksheet.cell(2 + row_index, 5, transform_type(attribute_datatype))

    wb = Workbook()
    ws = wb.active
    ws.title = TITLES[0]

    ad = create_definition(ws)
    create_zuweisung("bauteilKlassifikation", wb.create_sheet(TITLES[1]))  # Todo: Make bk variable
    create_internal_mapping(ad, wb.create_sheet(TITLES[2]))
    wb.save(path)



from __future__ import annotations
import SOMcreator
from SOMcreator import classes
from SOMcreator.filehandling import core
from SOMcreator.filehandling.constants import PREDEFINED_PSETS, ATTRIBUTES
from SOMcreator.filehandling import attribute
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from SOMcreator.filehandling.typing import PropertySetDict, MainDict
    from SOMcreator import Project


def load(proj: Project, pset_dict: PropertySetDict, identifier: str, obj: classes.Object | None) -> None:
    name, description, optional, parent, filter_matrix = core.get_basics(proj, pset_dict)
    pset = classes.PropertySet(name=name, obj=obj, uuid=identifier, description=description, optional=optional,
                               project=proj, filter_matrix=filter_matrix)
    attributes_dict = pset_dict[ATTRIBUTES]
    for ident, attribute_dict in attributes_dict.items():
        attribute.load(proj, attribute_dict, ident, pset)
    SOMcreator.filehandling.parent_dict[pset] = parent





#### Export ####

def write_entry(pset: classes.PropertySet) -> PropertySetDict:
    pset_dict: PropertySetDict = dict()
    core.write_basics(pset_dict, pset)
    attributes_dict = dict()
    for attrib in pset.get_all_attributes():
        new_dict = attribute.write(attrib)
        attributes_dict[attrib.uuid] = new_dict
    pset_dict[ATTRIBUTES] = attributes_dict
    return pset_dict


